module earth_example_dll
  use iso_c_binding, only: c_int, c_double, c_char, c_ptr, c_loc, c_f_pointer
  use iso_fortran_env, only: real32, real64

  implicit none  

  private

  ! Kind parameters which should be used when defining variables
  integer, parameter :: sp = real32  !< Single precision real
  integer, parameter :: dp = real64  !< Double precision real
  !
  !------------------------------------------------------------------------------------------
  ! linear soil spring example
  !------------------------------------------------------------------------------------------  
  type :: model_data_t
!     real(dp):: kspring
     integer :: io_inpfile
     integer :: io_errfile     
  end type
contains

  !----------------------------------------------------------------------------------------------------  
  ! USER_DEFINES_INIT
  !----------------------------------------------------------------------------------------------------  
  !DIR$ ATTRIBUTES DLLEXPORT :: userdef_files_init
  !----------------------------------------------------------------------------------------------------    
  subroutine userdef_files_init(instance, c_ch_usrelname,c_ch_eltyp,  &
                                c_ch_usrelinp,c_ch_elID, ierr) bind(c)
    !---------------------------------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------------------------------
    Implicit None    
    !  ------------------------------------------------------- External variables-----------------------
    !> pointer to memory for this dll
    type(c_ptr), intent(inout) :: instance
    !> ch_usrelname : name of the dll 
    character(kind=c_char), intent(in) :: c_ch_usrelname(*)
    !> ch_eltyp     : element type
    character(kind=c_char), intent(in) :: c_ch_eltyp(*)
    !> ch_usrelinp     : input file for the dll
    character(kind=c_char), intent(in) :: c_ch_usrelinp(*)
    !> ch_elID    : user defined ID
    character(kind=c_char), intent(in) :: c_ch_elID(*)    
    !> ierr   : Error flag
    integer(c_int), intent(out):: ierr
    !------------------------------------------------------- Internal variables-------------------------
    character (len=256) :: ch_errorFile, f_ch_usrelinp, f_ch_usrelname, f_ch_eltyp, f_ch_elID
    character (len=4)   :: ch_suffix
    character (len=20)  :: ch_subtext        
    type(model_data_t), pointer :: model_data
    !----------------------------------------------------------------------------------------------------
    ! -- Initialization
    !----------------------------------------------------------------------------------------------------
    ierr = 0
    ch_suffix=".log"
    ch_subtext="_test"        
    !----------------------------------------------------------------------------------------------------      
    ! convert c characters back to fortran characters
    !----------------------------------------------------------------------------------------------------      
    f_ch_usrelinp  = to_f_chars(c_ch_usrelinp)
    f_ch_usrelname = to_f_chars(c_ch_usrelname)
    f_ch_eltyp     = to_f_chars(c_ch_eltyp)
    f_ch_elID      = to_f_chars(c_ch_elID)            
    !----------------------------------------------------------------------------------------------------
    ! Make an error file for the specific model using the ID
    !----------------------------------------------------------------------------------------------------    
    ch_errorFile = trim(f_ch_elID)//trim(ch_subtext)//trim(ch_suffix)
    !----------------------------------------------------------------------------------------------------
    ! -- Main section    
    !----------------------------------------------------------------------------------------------------     
    allocate(model_data)                 ! allocate this model data
    instance = c_loc(model_data)         ! pointer to this model
!    
    open(file=ch_errorFile, newunit=model_data%io_errfile, iostat=ierr)
    if (ierr /=0 ) goto 7100
!
    open(file=trim(f_ch_usrelinp), newunit=model_data%io_inpfile, iostat=ierr)
    if (ierr /=0 ) goto 7200
    ! ....
    !  etc etc

    
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Error section
    !----------------------------------------------------------------------------------------------------  
7000 continue
    ierr = -1
    write(model_data%io_errfile,9000)trim(f_ch_usrelname)
    goto 8000
    !
7100 continue
    ierr = -2  
    write(model_data%io_errfile,9010)trim(ch_errorfile)
    goto 8000
    !    
7200 continue
    ierr = -3
    write(model_data%io_errfile,9020)trim(f_ch_usrelinp)
    goto 8000

    !----------------------------------------------------------------------------------------------------  
    ! -- Format statements
    !----------------------------------------------------------------------------------------------------      
    !
9000 format(5X,'ERROR calling user defined element :',A)            
9010 format(/4x,'*** ERROR open error file : ',A20//)
9020 format(/4x,'*** ERROR open input file : ',A20//)
    
8000 continue
  end subroutine userdef_files_init
  
  !----------------------------------------------------------------------------------------------------  
  ! USER_DEFINES_INIT
  !----------------------------------------------------------------------------------------------------  
  !DIR$ ATTRIBUTES DLLEXPORT :: userdef_element_init
  !----------------------------------------------------------------------------------------------------    
  subroutine userdef_element_init(instance, c_ch_usrelname,c_ch_eltyp,c_ch_usrelinp,c_ch_elID,  &
                                  iel,init_xyz_iel,init_incli_i,                      &       
                                  initial_depth_i,ldyn,ierr) bind(c)
    !---------------------------------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------------------------------
    Implicit None    
    !  ------------------------------------------------------- External variables-----------------------
    !> pointer to memory for this dll
    type(c_ptr), value, intent(in)     :: instance
    !> ch_usrelname : name of the dll 
    character(kind=c_char), intent(in) :: c_ch_usrelname(*)
    !> ch_eltyp     : element type
    character(kind=c_char), intent(in) :: c_ch_eltyp(*)
    !> ch_usrelinp     : input file for the dll
    character(kind=c_char), intent(in) :: c_ch_usrelinp(*)
    !> ch_elID    : user defined ID
    character(kind=c_char), intent(in) :: c_ch_elID(*)    
    !
    !> init_xyz_iel : initial coordinates x,y,z for element iel    
    real(c_double), dimension(3), intent(in) :: init_xyz_iel
    !> init_xyz_iel : initial inclination x,y,z [rad] for element iel
    !>                inclination is based on the elements reference vector
    real(c_double), intent(in) :: init_incli_i
    !> init_xyz_iel : initial depth x,y,z for element iel        
    real(c_double), intent(in) :: initial_depth_i
    !
    !> iel          : element number    
    integer(c_int), intent(in) :: iel
    !> ldyn         : control variable, 0: Static analysis
    !>                                  1: Linear dynamic            (illegal)
    !>                                  2: Nonlinear dynamic      
    !>                                  3: Frequency domain analysis (illegal)       
    integer(c_int), intent(in) :: ldyn
    !> ierr   : Error flag
    integer(c_int), intent(out):: ierr
    !------------------------------------------------------- Internal variables-------------------------
    character (len=256) :: ch_errorFile, f_ch_usrelinp, f_ch_usrelname, f_ch_eltyp, f_ch_elID  
    type(model_data_t), pointer :: model_data
    !----------------------------------------------------------------------------------------------------
    ! -- Initialization
    !----------------------------------------------------------------------------------------------------  
    ierr = 0
    call c_f_pointer(instance, model_data)    
    !----------------------------------------------------------------------------------------------------      
    ! convert c characters back to fortran characters
    !----------------------------------------------------------------------------------------------------      
    f_ch_usrelinp  = to_f_chars(c_ch_usrelinp)
    f_ch_usrelname = to_f_chars(c_ch_usrelname)
    f_ch_eltyp     = to_f_chars(c_ch_eltyp)
    f_ch_elID      = to_f_chars(c_ch_elID)            
    !----------------------------------------------------------------------------------------------------
    !
    !----------------------------------------------------------------------------------------------------    

    !----------------------------------------------------------------------------------------------------
    ! -- Main section
    !----------------------------------------------------------------------------------------------------  
    ! ....
    !  etc etc
    
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Error section
    !----------------------------------------------------------------------------------------------------  
7000 continue
    ierr = -1
    write(model_data%io_errfile,9000)trim(f_ch_usrelname)
    goto 8000
    !  
    !----------------------------------------------------------------------------------------------------  
    ! -- Format statements
    !----------------------------------------------------------------------------------------------------      
    !
9000 format(5X,'ERROR calling user defined element :',A)            
         
    
8000 continue
    !
  end subroutine userdef_element_init


  !----------------------------------------------------------------------------------------------------  
  !
  !----------------------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------------------  
  ! USER_DEFINES_FINISH
  !----------------------------------------------------------------------------------------------------  
  !DIR$ ATTRIBUTES DLLEXPORT :: userdef_element_calculate
  !----------------------------------------------------------------------------------------------------    
  subroutine userdef_element_calculate(instance, c_ch_usrelname,c_ch_elID, iel, diameter,            &
                                       elin, xyz_iel, depth_iel,                                     &
                                       incli_iel, tracon_iel, vel_iel,iter,time,rts ,                &
                                       ldyn, usrfor, usrstf, usrdamp,ierr) bind(c)
    !---------------------------------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------------------------------
    implicit none
    !------------------------------------------------------- External variables-------------------------
    !> pointer to memory for this dll
    type(c_ptr), value, intent(in) :: instance
    !> ch_usrelname : name of the dll 
    character(kind=c_char), intent(in) :: c_ch_usrelname(*)
    !> name of user defined id
    character(kind=c_char), intent(in) :: c_ch_elID(*)    
    !> iel          : element number    
    integer(c_int), intent(in) :: iel
    !> diameter    :  external contact diameter 
    real(c_double), intent(in) :: diameter
    !> elin : element length
    !> Note: the length is equal to the half-length between the nodes on the reference element
    real(c_double), intent(in) :: elin    
    !> xyz_iel : initial coordinates x,y,z for element iel    
    real(c_double), dimension(3), intent(in) :: xyz_iel
    !> depth_i  : initial depth x,y,z for element iel        
    real(c_double), intent(in) :: depth_iel
    !> incli_iel : initial inclination x,y,z [rad] for element iel
    !>                inclination is based on the elements reference vector
    real(c_double), intent(in) :: incli_iel
    !> Transformation matrix between local and global systems
    !> vlocal= tracon * vglobal 
    real(c_double), dimension(3,3), intent(in) :: tracon_iel
    real(c_double), dimension(3), intent(in) :: vel_iel
    !> usrfor :  forces in x, y & z-direction (global)
    real(c_double), dimension(3), intent(out)   :: usrfor
    !> usrstf  : tangent stiffness matrix (global)
    real(c_double), dimension(3,3), intent(out) :: usrstf
    !> usrdamp :  damping matrix (global)
    real(c_double), dimension(3,3), intent(out) :: usrdamp
    !> iter : iteration step
    integer(c_int), intent(in) :: iter
    !> time : Accumulated time[s]
    real(c_double), intent(in) :: time
    !> rts  : time step
    real(c_double), intent(in) :: rts        

    !> ldyn         : control variable, 0: Static analysis
    !>                                  1: Linear dynamic            (illegal)
    !>                                  2: Nonlinear dynamic      
    !>                                  3: Frequency domain analysis (illegal)       
    integer(c_int), intent(in) :: ldyn    
    !> ierr   : Error flag
    integer(c_int), intent(out):: ierr   
    !------------------------------------------------------- Internal variables-------------------------
    character (len=256) :: f_ch_usrelname, f_ch_elID
    !
    real(dp) :: cmp_depth,dz
    type(model_data_t), pointer :: model_data
    !----------------------------------------------------------------------------------------------------
    ! -- Initialization
    !----------------------------------------------------------------------------------------------------  
    ierr = 0
    call c_f_pointer(instance, model_data)
    !----------------------------------------------------------------------------------------------------          
    ! convert c characters back to fortran characters
    !----------------------------------------------------------------------------------------------------      
    f_ch_usrelname = to_f_chars(c_ch_usrelname)
    f_ch_elID      = to_f_chars(c_ch_elID)        
    !----------------------------------------------------------------------------------------------------
    ! -- Main section
    !----------------------------------------------------------------------------------------------------  
    ! ....
    !  etc etc

    
    !----------------------------------------------------------------------------------------------------
    ! --
    !----------------------------------------------------------------------------------------------------  
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Error section
    !----------------------------------------------------------------------------------------------------  
7000 continue
    write(model_data%io_errfile,9000)trim(f_ch_usrelname)

    !----------------------------------------------------------------------------------------------------  
    ! -- Format statements
    !----------------------------------------------------------------------------------------------------      
    !
9000 format(5X,'ERROR calling user defined element :',A)            

8000 continue
    
  end subroutine userdef_element_calculate

  !----------------------------------------------------------------------------------------------------  
  ! USER_DEFINES_FINISH
  !----------------------------------------------------------------------------------------------------  
  !DIR$ ATTRIBUTES DLLEXPORT :: userdef_element_finish
  !----------------------------------------------------------------------------------------------------    
  subroutine userdef_element_finish(instance, c_ch_usrelname,c_ch_eltyp, &
                                    c_ch_usrelinp,c_ch_elID, iel, ierr) bind(c)
    !---------------------------------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------------------------------
    implicit none
    !------------------------------------------------------- External variables-------------------------
    !> pointer to memory for this dll
    type(c_ptr), value, intent(in) :: instance
    !> ch_usrelname : name of the dll 
    character(kind=c_char), intent(in) :: c_ch_usrelname(*)
    !> ch_eltyp     : element type
    character(kind=c_char), intent(in) :: c_ch_eltyp(*)
    !> ch_usrelinp     : input file for the dll
    character(kind=c_char), intent(in) :: c_ch_usrelinp(*)
    !> ch_elID    : user defined ID
    character(kind=c_char), intent(in) :: c_ch_elID(*)       
    !> iel          : element number    
    integer(c_int), intent(in) :: iel
    !> ierr   : Error flag
    integer(c_int), intent(out):: ierr   
    !------------------------------------------------------- Internal variables-------------------------
    character (len=256) :: f_ch_usrelinp, f_ch_usrelname, f_ch_eltyp , f_ch_elID
    type(model_data_t), pointer :: model_data
    !----------------------------------------------------------------------------------------------------
    ! -- Initialization
    !----------------------------------------------------------------------------------------------------  
    ierr = 0
    call c_f_pointer(instance, model_data)
    !----------------------------------------------------------------------------------------------------      
    ! convert c characters back to fortran characters
    !----------------------------------------------------------------------------------------------------      
    f_ch_usrelinp  = to_f_chars(c_ch_usrelinp)
    f_ch_usrelname = to_f_chars(c_ch_usrelname)
    f_ch_eltyp     = to_f_chars(c_ch_eltyp)
    f_ch_elID      = to_f_chars(c_ch_elID)           
    !----------------------------------------------------------------------------------------------------
    ! -- Main section
    !----------------------------------------------------------------------------------------------------  
    ! ....
    !  etc etc

    
    !----------------------------------------------------------------------------------------------------  
    ! Close files
    !----------------------------------------------------------------------------------------------------      

    
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Error section
    !----------------------------------------------------------------------------------------------------  
7000 continue
    write(model_data%io_errfile,9000)trim(f_ch_usrelname)
    !
7100 continue
    write(model_data%io_errfile,9010)
    goto 8000
    !    
7200 continue
    write(model_data%io_errfile,9020)
    !
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Format statements
    !----------------------------------------------------------------------------------------------------      
    !
9000 format(5X,'ERROR calling user defined element :',A)            
9010 format(/4x,'*** ERROR closing error file ')
9020 format(/4x,'*** ERROR closing input file ')            
    !

8000 continue
  end subroutine userdef_element_finish
  !----------------------------------------------------------------------------------------------------  
  !
  !----------------------------------------------------------------------------------------------------        
  !----------------------------------------------------------------------------------------------------  
  ! USER_DEFINES_FINISH
  !----------------------------------------------------------------------------------------------------  
  !DIR$ ATTRIBUTES DLLEXPORT :: userdef_files_finish
  !----------------------------------------------------------------------------------------------------    
  subroutine userdef_files_finish(instance, c_ch_usrelname,c_ch_eltyp, &
                                    c_ch_usrelinp,c_ch_elID,ierr) bind(c)
    !---------------------------------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------------------------------
    implicit none
    !------------------------------------------------------- External variables-------------------------
    !> pointer to memory for this dll
    type(c_ptr), value, intent(in) :: instance
    !> ch_usrelname : name of the dll 
    character(kind=c_char), intent(in) :: c_ch_usrelname(*)
    !> ch_eltyp     : element type
    character(kind=c_char), intent(in) :: c_ch_eltyp(*)
    !> ch_usrelinp     : input file for the dll
    character(kind=c_char), intent(in) :: c_ch_usrelinp(*)
    !> ch_elID    : user defined ID
    character(kind=c_char), intent(in) :: c_ch_elID(*)       
    !> ierr   : Error flag
    integer(c_int), intent(out):: ierr   
    !------------------------------------------------------- Internal variables-------------------------
    character (len=256) :: f_ch_usrelinp, f_ch_usrelname, f_ch_eltyp , f_ch_elID
    type(model_data_t), pointer :: model_data
    !----------------------------------------------------------------------------------------------------
    ! -- Initialization
    !----------------------------------------------------------------------------------------------------  
    ierr = 0
    call c_f_pointer(instance, model_data)
    !----------------------------------------------------------------------------------------------------      
    ! convert c characters back to fortran characters
    !----------------------------------------------------------------------------------------------------      
    f_ch_usrelinp  = to_f_chars(c_ch_usrelinp)
    f_ch_usrelname = to_f_chars(c_ch_usrelname)
    f_ch_eltyp     = to_f_chars(c_ch_eltyp)
    f_ch_elID      = to_f_chars(c_ch_elID)           
    !----------------------------------------------------------------------------------------------------
    ! -- Main section
    !----------------------------------------------------------------------------------------------------  
    ! ....
    !  etc etc




    
    !----------------------------------------------------------------------------------------------------  
    ! Close files
    !----------------------------------------------------------------------------------------------------      
       
    close(model_data%io_errfile, iostat=ierr)
    if (ierr /=0 ) goto 7100       

    deallocate(model_data)              
    
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Error section
    !----------------------------------------------------------------------------------------------------  
7000 continue
    write(model_data%io_errfile,9000)trim(f_ch_usrelname)
    !
7100 continue
    write(model_data%io_errfile,9010)
    goto 8000
    !    
7200 continue
    write(model_data%io_errfile,9020)
    !
    goto 8000
    !----------------------------------------------------------------------------------------------------  
    ! -- Format statements
    !----------------------------------------------------------------------------------------------------      
    !
9000 format(5X,'ERROR calling user defined element :',A)            
9010 format(/4x,'*** ERROR closing error file ')
9020 format(/4x,'*** ERROR closing input file ')            
    !

8000 continue
  end subroutine userdef_files_finish

  pure function to_f_chars(cchars) result(chars)
    !----------------------------------------------------------------------------------------------------
    ! convert c characters to fortran characters
    !----------------------------------------------------------------------------------------------------        
    use iso_c_binding, only: c_char, c_null_char
    implicit none
    character(len=1, kind=c_char), intent(in) :: cchars(*)
    character(len=:), allocatable :: chars
    integer :: i, n
    character :: c
    ! Find length
    i = 1
    c = cchars(i)
    do while(c /= c_null_char)
       i = i + 1
       c = cchars(i)
    end do
    n = i - 1
    allocate(character(len=n) :: chars)
    do i = 1, n
       chars(i:i) = cchars(i)
    end do
  end function to_f_chars

end module earth_example_dll
